from rich.text import Text
from textual.app import ComposeResult
from textual.containers import Center
from textual.message import Message
from textual.reactive import var
from textual.widget import Widget
from textual.widgets import DataTable, Label


class TransactionHistory(Widget):
    HEADER = ("Out/In", "Date", "Time", "Value", "Units")
    DEFAULT_CSS = """
    TransactionHistory {
        height: 100%;
        width: 1fr;
        margin-top: 1;
        margin-bottom: 1;
        background: $panel;
        outline-top: tall $primary;
        outline-left: tall $primary;
        outline-right: tall $primary;
    }
    .top {
        outline-top: tall $primary;
        width: 100%;
        height: 3;
        margin: 0 1 0 1;
        content-align: center middle;
        background: $panel;
    }
    .bottom {
        outline-bottom: tall $primary;
        width: 100%;
        height: 2;
        margin: 0 1 0 1;
        padding-bottom: 1;
        background: $panel;
    }
    TransactionHistory>Center {
        margin-left: 1;
        overflow: auto;
        height: 1fr;
        scrollbar-gutter: stable;

    }
    TransactionHistory>Center>DataTable {
        background: $panel;
        height: auto;
        width: auto;
    }
    """

    rows = var([])

    class TxOverlayRequested(Message):
        def __init__(self, txid: str):
            self.txid = txid
            super().__init__()

    def __init__(self, rows: list[tuple[str | Text]]) -> None:
        self._rows = rows
        self.selected_row: str | None = None
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Label("Previous Transactions", classes="top")
        yield Center(DataTable())
        yield Label("", classes="bottom")

    def on_mount(self) -> None:
        table = self.query_one("DataTable", DataTable)

        for col in self.HEADER:
            table.add_column(col, key=col)

        self.rows = self._rows
        table.cursor_type = "row"

    def on_data_table_row_selected(self, event: DataTable.RowSelected):
        print("ROW SELECTED")
        event.stop()

        if not event.row_key.value:
            return

        if self.selected_row != event.row_key.value:
            self.selected_row = event.row_key.value
            return

        self.post_message(self.TxOverlayRequested(event.row_key.value))

        # could add a timing element here? Like 100ms?
        # send message with txdata to parent

    def watch_rows(self, new_value: list[tuple]):
        table = self.query_one("DataTable", DataTable)

        self._rows = new_value

        if not self._rows:
            table.clear()

        for row in self._rows:
            table.add_row(*row[1:], key=row[0])

        table.sort("Date", "Time", reverse=True)
        # table.cursor_coordinate = Coordinate(2, 0)



from textual.app import ComposeResult
from textual.containers import Grid
from textual.screen import ModalScreen
from textual.widgets import Button, Label


class AddressBookOverlay(ModalScreen[bool]):
    DEFAULT_CSS = """
        AddressBookOverlay {
            align: center middle;
        }

        #dialog {
            grid-size: 2;
            grid-gutter: 1 2;
            grid-rows: 1fr 3;
            padding: 0 1;
            width: 60;
            height: 11;
            border: thick $background 80%;
            background: $surface;
        }

        #question {
            column-span: 2;
            height: 1fr;
            width: 1fr;
            content-align: center middle;
        }

        AddressBookOverlay>Grid>Button {
            width: 100%;
        }"""

    def __init__(self, nickname: str, address: str) -> None:
        self.nickname = nickname
        self.address = address

        super().__init__()

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(
                f"Are you sure you want to delete address book entry {self.nickname} with address: {self.address}",
                id="question",
            ),
            Button("Delete", variant="error", id="delete"),
            Button("Cancel", variant="primary", id="cancel"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "delete":
            delete = True
        else:
            delete = False

        self.dismiss(delete)



import re

from textual import on
from textual.app import ComposeResult
from textual.containers import Center, Container, Horizontal
from textual.message import Message
from textual.reactive import var
from textual.screen import Screen
from textual.validation import Function
from textual.widgets import Button, Input, Select, Static


class ImportKeyToWallet(Screen[tuple[str, str] | None]):
    class KeyImportRequested(Message):
        def __init__(self, key: str) -> None:
            self.key = key
            super().__init__()

    DEFAULT_CSS = """
    """
    private_key = var("")

    def __init__(
        self, wallet_name: str, wallets: list, network_name: str, networks: list
    ):
        self.wallet_name = wallet_name
        self.wallets = wallets
        self.network_name = network_name
        self.networks = networks
        super().__init__()

    def on_mount(self) -> None:
        private_key = self.query_one("Input", Input)
        self.set_focus(private_key)

    def input_pk_validate(self, pk: str) -> bool:
        if pk == "":
            return True

        length = len(pk)
        regex = r"^[5KL][1-9A-HJ-NP-Za-km-z]{50,51}$"

        if length == 51 or length == 52:
            return bool(re.search(regex, pk))
        elif length == 64:
            try:
                int(pk, 16)
            except ValueError:
                return False
            else:
                return True
        else:
            return False

    def compose(self) -> ComposeResult:
        yield Horizontal(
            Container(Button("X", variant="error", id="exit"), id="button_container"),
            Center(Static("Import Private Key"), id="import_key_title"),
            id="import_key_top_bar",
        )
        yield Horizontal(
            Input(
                None,
                placeholder="Paste or type your private key in WIF or Hex format",
                id="input_private_key",
                password=True,
                validators=[
                    Function(function=self.input_pk_validate),
                ],
            ),
            Button("\U0001F441", id="import_key_toggle"),
            id="private_key_container",
        )
        yield Horizontal(
            Select(
                id="wallet_id",
                options=[(x, x) for x in self.wallets],
                value=self.wallet_name,
                classes="import_key_select",
            ),
            Select(
                id="network_id",
                options=[(x, x) for x in self.networks],
                value=self.network_name,
                classes="import_key_select",
            ),
            id="import_key_dropdowns",
        )
        yield Horizontal(
            Button(
                "Clear",
                variant="warning",
                id="clear_private_key",
                classes="import_key_buttons",
            ),
            Button("Import", id="import_key_button", classes="import_key_buttons"),
            id="import_key_buttons_container",
        )

    @on(Input.Changed, "#input_private_key")
    def address_input_changed(self, event: Input.Changed) -> None:
        event.stop()

        if not event.validation_result.failures:
            self.private_key = event.value
        else:
            self.private_key = ""

    @on(Button.Pressed)
    def on_button_pressed(self, event: Button.Pressed) -> None:
        event.stop()

        if event.button.id == "import_key_toggle":
            self.toggle_key_view()

        if event.button.id == "clear_private_key":
            private_key = self.query_one("Input", Input)
            private_key.value = ""
            self.set_focus(private_key)
        if event.button.id == "import_key_button":
            wallet_select = self.query_one("#wallet_id", Select)
            network_select = self.query_one("#network_id", Select)
            self.dismiss((wallet_select.value, network_select.value, self.private_key))

        if event.button.id == "exit":
            self.dismiss()

    def toggle_key_view(self):
        private_key = self.query_one("Input", Input)

        if private_key.password:
            private_key.password = False
        else:
            private_key.password = True

    def watch_private_key(self, new_value: str) -> None:
        submit = self.query_one("#import_key_button", Button)

        if new_value:
            submit.disabled = False
        else:
            submit.disabled = True



import pyperclip
from fluxwallet.keys import HDKey
from fluxwallet.mnemonic import Mnemonic
from fluxwallet.wallet import Wallet
from textual import work
from textual.app import ComposeResult
from textual.containers import Center, Container, Grid, Horizontal
from textual.message import Message
from textual.screen import ModalScreen
from textual.widgets import Button, Static

import asyncio

from lucidwallet.helpers import init_app
from lucidwallet.screens import WalletLanding

MNEMONIC_TEXT = """Here is your new mnemonic. You will only get to see this once.
Write down, or copy to clipboard if you like to YOLO."""


class Notification(Static):
    def on_mount(self) -> None:
        self.set_timer(3, self.remove)

    def on_click(self) -> None:
        self.remove()


class MnemonicOverlay(ModalScreen[bool]):
    DEFAULT_CSS = """

    """

    class WalletCreated(Message):
        def __init__(self, wallet: Wallet) -> None:
            self.wallet = wallet

            super().__init__()

    def __init__(self, nickname: str, mnemonic: str) -> None:
        self.nickname = nickname
        self.mnemonic = mnemonic
        self.words = mnemonic.split(" ")

        self.hidden_words = [
            f"{index}: {'*' * len(word)}" for index, word in enumerate(self.words, 1)
        ]
        self.hidden_dom = [
            Static(f"{index}: {'*' * len(word)}")
            for index, word in enumerate(self.words, 1)
        ]

        super().__init__()

    def compose(self) -> ComposeResult:
        yield Container(
            Container(
                Button("X", variant="error", id="exit"),
                id="mnemonic_exit",
            ),
            Center(Static(MNEMONIC_TEXT, id="mnemonic_notice")),
            Grid(*self.hidden_dom),
            Horizontal(
                Button("Reveal", variant="success", id="reveal"),
                Button("Copy", variant="warning", id="copy"),
                id="mnemonic_button_container",
            ),
        )

    async def on_mount(self) -> None:
        # self.query_one("#reveal", Button).focus()
        self.create_wallet()

    @work()
    async def create_wallet(self):
        # have already validated, but validate again
        seed = Mnemonic().to_seed(self.mnemonic).hex()
        hdkey = HDKey.from_seed(seed, network="flux")

        wallet = await Wallet.create(
            self.nickname,
            hdkey,
            network="flux",
        )

        if wallet:
            await wallet.new_key(network="bitcoin")
            app_data = await init_app()

            if not self.app.is_screen_installed("wallet_landing"):
                self.app.install_screen(
                    WalletLanding(
                        wallet,
                        app_data.networks,
                        app_data.wallets,
                    ),
                    name="wallet_landing",
                )

            self.post_message(self.WalletCreated(wallet))

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "reveal":
            dom_words = self.query("Grid>Static")
            if str(event.button.label) == "Reveal":
                event.button.label = "Hide"

                for index, (dom_item, word) in enumerate(zip(dom_words, self.words), 1):
                    dom_item.update(f"{index}: {word}")

            elif str(event.button.label) == "Hide":
                event.button.label = "Reveal"
                for dom_word, word in zip(dom_words, self.hidden_words):
                    dom_word.update(word)

        elif event.button.id == "copy":
            word_str = " ".join(self.words)

            try:
                self.app.config.copy_callback(word_str)
            except Exception:  # Fix
                self.notify("No clipboard available", severity="warning")
            else:
                self.notify("Mnemonic copied!")
        else:
            # this is ugly, only seems to be docker that's slow??
            while not self.app.is_screen_installed("wallet_landing"):
                await asyncio.sleep(0.05)

            self.app.switch_screen("wallet_landing")



import asyncio

import pyperclip
from bitcoin.signmessage import BitcoinMessage, SignMessage
from bitcoin.wallet import CBitcoinSecret, P2PKHBitcoinAddress
from rich.text import Text
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Center, Container, Horizontal
from textual.message import Message
from textual.screen import Screen
from textual.widgets import Button, Input, Label, Static


# class Notification(Static):
#     def on_mount(self) -> None:
#         self.set_timer(3, self.remove)

#     def on_click(self) -> None:
#         self.remove()


class CopyLabel(Label):
    def action_copy_clipboard(self, text: str):
        try:
            self.app.config.copy_callback(text)
        except Exception:  # Fix
            self.notify("No clipboard available", severity="warning")
        else:
            self.notify("Copied")

    # class Copied(Message):
    #     ...


class SignMessageOverlay(Screen):
    DEFAULT_CSS = """

    """

    BINDINGS = [
        ("ctrl+t", "app.toggle_dark", "Toggle Dark mode"),
        Binding("ctrl+c,ctrl+q", "app.quit", "Quit"),
        (
            "escape",
            "app.pop_screen()",
            "home",
        ),
    ]

    def __init__(self, private_key: bytes) -> None:
        self.key = CBitcoinSecret.from_secret_bytes(private_key)
        self.address = P2PKHBitcoinAddress.from_pubkey(self.key.pub)

        super().__init__()

    def compose(self) -> ComposeResult:
        yield Horizontal(
            Container(Button("X", variant="error", id="exit"), id="button_container"),
            Center(Static("SIGN MESSAGE"), id="title"),
            id="sign_top_bar",
        )

        yield Input("", placeholder="Enter Message", id="sign_message_input")
        yield Center(
            CopyLabel(
                f"Address: [@click=copy_clipboard('{self.address}')]{self.address}[/]"
            ),
            id="sign_message_address",
        )
        yield Horizontal(
            Button("Clear", variant="warning", id="sign_message_clear"),
            Button("Sign", variant="success", id="sign_message_submit"),
            id="sign_buttons",
        )

    def on_mount(self) -> None:
        message = self.query_one("Input", Input)
        self.set_focus(message)

    def on_button_pressed(self, event: Button.Pressed) -> None:
        message = self.query_one("Input", Input)
        if event.button.id == "sign_message_submit" and message.value:
            sig_value = self.sign_message(message.value)
            sig = self.query("#sign_message_sig")
            if not sig:
                self.mount(
                    Center(
                        CopyLabel(
                            f"Signature: [@click=copy_clipboard('{sig_value}')]{sig_value}[/]"
                        ),
                        id="sign_message_sig",
                    )
                )

        if event.button.id == "sign_message_clear":
            message.value = ""
            sigs = self.query("#sign_message_sig")
            if sigs:
                sigs.last().remove()

            self.set_focus(message)

        if event.button.id == "exit":
            self.dismiss()

    def sign_message(self, msg: str) -> str:
        message = BitcoinMessage(msg)

        return SignMessage(self.key, message).decode()

    # def on_copy_label_copied(self):
    #     self.mount(Notification(Text("Copied!")))


if __name__ == "__main__":

    class BlahApp(App):
        def on_mount(self):
            self.push_screen(SignMessageOverlay())

    blah = BlahApp(css_path="/Users/davew/code/flux/wallet_app/app.css")
    blah.run()



from functools import partial
from typing import Iterable

from fluxwallet.db import Db, DbAddressBook
from rich.text import Text
from textual import on
from textual.app import App, ComposeResult
from textual.containers import Center, Horizontal
from textual.message import Message
from textual.reactive import var
from textual.screen import Screen
from textual.validation import Function, Regex
from textual.widgets import Button, DataTable, Input

from lucidwallet.screens import AddressBookOverlay


class AddressBook(Screen[str | None]):
    DEFAULT_CSS = """
    AddressBook {
        width: 1fr;
        align: center top;
        border: double $primary;
        padding: 2;
        margin: 4;
        border-title-align: center;
        border-title-color: $secondary;

    }
    AddressBook > Input {
        margin-bottom: 1;
    }

    AddressBook > Horizontal {
        margin: 1;
        height: auto;
        align: center top;
    }
        AddressBook > Horizontal > Button {
        margin: 1;
    }

    AddressBook>Center>DataTable {
        background: $primary-background;
        width: auto;
    }
    """

    address = var("")

    def __init__(self, address: str | None = None, return_nickname: bool = False):
        self._address = address

        self.return_nickname = return_nickname
        self.table = None

        super().__init__()

    def on_mount(self):
        self.border_title = "Address book"
        self.address = self._address

        self.table = self.query_one("DataTable", DataTable)

        for col in ["Nickname", "Address", "Remove"]:
            self.table.add_column(col)

        self.session = Db().session
        self.fetch_address_book()

        # table.cursor_type = "row"

    def on_unmount(self):
        self.session.close()

    def fetch_address_book(self):
        address_book: Iterable[DbAddressBook] = self.session.query(DbAddressBook).all()

        for row in address_book:
            print(
                "ROW",
                repr(
                    self.table.add_row(
                        row.name,
                        row.address,
                        Text("X", style="bold red", justify="center"),
                    )
                ),
            )

    def validate_address_input(self, address: str) -> bool:
        if address == "":
            return True

        if address.startswith(("t1", "t3")) and len(address) == 35:
            self.sendto_address = address
            return True
        return False

    def compose(self) -> ComposeResult:
        yield Input(
            None,
            validators=[Regex("^[a-zA-Z0-9_-]*$")],
            placeholder="Address nickname. Must only contain: a-z A-Z 0-9, underscore or hyphen",
            id="nickname",
        )
        yield Input(
            self._address,
            validators=[
                Function(function=self.validate_address_input),
            ],
            placeholder="Address",
            id="address",
        )
        yield Horizontal(
            Button("Save", id="address_save"),
            Button("Cancel", variant="error", id="cancel"),
        )
        yield Center(DataTable())

    @on(Input.Changed, "#address")
    def address_input_changed(self, event: Input.Changed) -> None:
        event.stop()

        if not event.validation_result.failures:
            self.addresxs = event.value
        else:
            self.address = ""

    @on(Button.Pressed)
    def on_button_pressed(self, event: Button.Pressed) -> None:
        event.stop()

        nickname = None

        if event.button.id == "address_save":
            nickname = self.query_one("#nickname", Input).value
            self.save_nickname(nickname, self.address)
            self.table.add_row(
                nickname, self.address, Text("X", justify="center", style="bold red")
            )
            # clear values

            if not self.return_nickname:
                nickname = None
        else:
            self.dismiss(nickname)

    def save_nickname(self, nickname: str, address: str):
        if not nickname:
            return

        address_book_entry: DbAddressBook = (
            self.session.query(DbAddressBook).filter_by(name=nickname).scalar()
        )

        if not address_book_entry:
            new_ab_entry = DbAddressBook(
                name=nickname,
                address=address,
            )
            self.session.add(new_ab_entry)
        else:
            address_book_entry.address = address

        try:
            self.session.commit()
        except Exception:
            self.session.rollback()
            raise

    def on_input_submitted(self):
        self.focus_next()

    def watch_address(self, new_value: bool) -> None:
        button = self.query_one("#address_save", Button)

        if new_value:
            button.disabled = False
            button.variant = "success"
        else:
            button.disabled = True
            button.variant = "default"

    def on_delete_row_callback(self, table: DataTable, row_key, delete: bool) -> None:
        if delete:
            nickname, _, _ = table.get_row(row_key)
            table.remove_row(row_key)
            self.session.query(DbAddressBook).filter_by(name=nickname).delete()
            self.session.commit()

    def on_data_table_cell_selected(self, event: DataTable.CellSelected):
        if str(event.value) == "X":
            table = self.query_one("DataTable", DataTable)
            nickname, address, _ = table.get_row(event.cell_key.row_key)
            callback = partial(
                self.on_delete_row_callback, table, event.cell_key.row_key
            )
            self.app.push_screen(
                AddressBookOverlay(nickname, address), callback=callback
            )



# overlay before address_book
from .address_book_overlay import AddressBookOverlay
from .address_book import AddressBook


# txsentoverlay before wallet_landing
# txoverlay before wallet_landing
# sign_message_overlay before wallet_landiong
# send_tx_overlay before wallet_landing
# import_key_to_wallet before wallet_landing
# wallet_landing before mnemonic_overlay
# mnemonic_overlay before create_wallet
from .tx_sent_overlay import TxSentOverlay
from .txoverlay import TxOverlay
from .sign_message_overlay import SignMessageOverlay
from .sendtx_overlay import SendTxOverlay
from .import_key_to_wallet import ImportKeyToWallet
from .wallet_landing import WalletLanding
from .get_encryption_password import EncryptionPassword
from .loading import LoadingScreen
from .mnemonic_overlay import MnemonicOverlay
from .create_wallet import CreateWallet


from .import_from_mnemonic import ImportFromMnemonic

from .keychain_overlay import KeychainOverlay

from .no_wallets import FirstRun



from textual import on
from textual.app import ComposeResult
from textual.containers import Center, Horizontal
from textual.reactive import var
from textual.screen import Screen
from textual.validation import Length
from textual.widgets import Button, Input, Static, Switch


class EncryptionPassword(Screen[tuple[str, bool]]):
    DEFAULT_CSS = """
        EncryptionPassword {
            layers: base notification;
        }
    """

    password = var("")

    def __init__(self, message: str | None = None) -> None:
        self.message = message
        super().__init__()

    def on_mount(self) -> None:
        password = self.query_one("#input_encryption_password", Input)
        matcher = self.query_one("#password_matcher", Static)
        matcher.visible = False
        keychain_switch = self.query_one("Switch", Switch)

        if not self.app.config.keyring_available:
            keychain_switch.disabled = True
            keychain_switch.tooltip = (
                "No keychain found. See https://pypi.org/project/keyring"
            )

        if self.message:
            self.notify(self.message)

        self.set_focus(password)

    def compose(self) -> ComposeResult:
        yield Horizontal(
            # Container(Button("X", variant="error", id="exit"), id="button_container"),
            Center(
                Static("Database Encryption password"),
                id="get_encryption_password_title",
            ),
            id="encryption_password_top_bar",
        )
        yield Horizontal(
            Input(
                None,
                placeholder="Paste or type your encryption password here",
                id="input_encryption_password",
                password=True,
                validators=Length(minimum=3),
            ),
            Button(
                "\U0001F441", id="show_password_toggle", classes="encryption_toggle"
            ),
        )
        yield Horizontal(
            Input(
                None,
                placeholder="Confirm your encryption password here",
                id="repeat_encryption_password",
                password=True,
                validators=Length(minimum=3),
            ),
            Button(
                "\U0001F441", id="repeat_password_toggle", classes="encryption_toggle"
            ),
        )
        yield Center(Static("Passwords do not match", id="password_matcher"))
        yield Horizontal(
            Switch(value=False),
            Static("Store in Keychain", classes="label"),
            id="switch_container",
        )
        yield Horizontal(
            Button("Clear", variant="warning", id="clear_password"),
            Button("Set Password", id="set_password", disabled=True),
            id="get_encryption_buttons",
        )

    @on(Input.Changed, "#input_encryption_password")
    def password_input_changed(self, event: Input.Changed) -> None:
        event.stop()

        matcher = self.query_one("#password_matcher", Static)
        repeat = self.query_one("#repeat_encryption_password", Input)

        if event.value != repeat.value:
            matcher.visible = True
            self.password = ""
            matcher.update("Passwords do not match")
            if matcher.has_class("-matcher_success"):
                matcher.remove_class("-matcher_success")
        elif not event.value:
            matcher.visible = False
            self.password = ""
        else:
            self.password = event.value
            matcher.update("Passwords match")
            matcher.add_class("-matcher_success")

    @on(Input.Changed, "#repeat_encryption_password")
    def repeat_input_changed(self, event: Input.Changed) -> None:
        event.stop()

        matcher = self.query_one("#password_matcher", Static)
        input_password = self.query_one("#input_encryption_password", Input)

        if event.value != input_password.value:
            self.password = ""
            matcher.visible = True
            matcher.update("Passwords do not match")
            if matcher.has_class("-matcher_success"):
                matcher.remove_class("-matcher_success")
        else:
            matcher.update("Passwords match")
            matcher.add_class("-matcher_success")
            if not event.validation_result.failures:
                self.password = event.value

    @on(Button.Pressed)
    def on_button_pressed(self, event: Button.Pressed) -> None:
        event.stop()

        if event.button.id == "show_password_toggle":
            to_toggle = self.query_one("#input_encryption_password", Input)
            self.toggle_password_view(to_toggle)
        if event.button.id == "repeat_password_toggle":
            to_toggle = self.query_one("#repeat_encryption_password", Input)
            self.toggle_password_view(to_toggle)

        if event.button.id == "clear_password":
            password = self.query_one("#input_encryption_password", Input)
            repeat_password = self.query_one("#repeat_encryption_password", Input)
            password.value = ""
            repeat_password.value = ""
            self.set_focus(password)

        if event.button.id == "set_password":
            password = self.query_one("#input_encryption_password", Input)
            # self.app_info.last_used_wallet.db.set_encrypted_key(password.value)
            store_in_keychain = self.query_one("Switch", Switch)
            self.dismiss((password.value, store_in_keychain.value))
            # self.app.switch_screen(
            #     WalletLanding(
            #         self.app_info.last_used_wallet,
            #         self.app_info.wallets,
            #         self.app_info.networks,
            #     )
            # )

        # if event.button.id == "exit":
        #     self.dismiss(None)

    def toggle_password_view(self, input: Input):
        if input.password:
            input.password = False
        else:
            input.password = True

    def watch_password(self, new_value: str) -> None:
        print("watch password")
        print("new value", new_value)
        submit = self.query_one("#set_password", Button)

        if new_value:
            submit.disabled = False
        else:
            submit.disabled = True



import hashlib

import keyring
from fluxwallet.db_new import Db
from textual.app import ComposeResult
from textual.containers import Center, Horizontal
from textual.screen import Screen
from textual.widgets import Button, Label

from lucidwallet.screens import EncryptionPassword

MESSAGE = """
Welcome to LucidWallet! This looks like your first time here... lets get you setup.

* BIP39 and BIP44 compliant.
* Supports multiple currencies. (Soon)

"""


class FirstRun(Screen):
    BINDINGS = [
        (
            "a",
            "add_weiner()",
            "Add Weiner",
        ),
    ]
    AUTO_FOCUS = ""

    def compose(self) -> ComposeResult:
        yield Center(
            Label(
                MESSAGE,
                id="first_time",
            )
        )
        yield Center(
            Label(
                "Password encrypt database? (You can change this later)",
                id="encrypt_db_label",
            ),
            id="label_container",
        )
        yield Horizontal(
            Button("Yes", id="encrypt_db", variant="primary"),
            Button("Skip", id="no_encrypt_db", variant="warning"),
            classes="horizontal_center",
        )
        yield Horizontal(
            Button(
                "Create Wallet", id="create_wallet", variant="primary", disabled=True
            ),
            Button(
                "Import from Mnemonic",
                id="from_mnemonic",
                variant="primary",
                disabled=True,
            ),
            Button(
                "Import from Private key (xpriv)",
                id="from_xpriv",
                variant="primary",
                disabled=True,
            ),
            # Button(
            #     "Set Db URI",
            #     id="select_database",
            #     variant="primary",
            # ),
            classes="horizontal_center",
        )
        # yield Footer()

    def keychain_callback(self, password: str, store_in_chain: bool) -> None:
        if not store_in_chain:
            return

        keyring.set_password("fluxwallet", "fluxwallet_user", password)

    async def on_encryption_password_set(self, result: tuple[str, bool]):
        password, store_in_keychain = result

        hashed = hashlib.sha256(bytes(password, "utf8")).hexdigest()

        # callback = partial(self.keychain_callback, hashed)
        # self.app.push_screen(KeychainOverlay(), callback)

        if store_in_keychain:
            keyring.set_password("fluxwallet", "fluxwallet_user", hashed)

        db = Db()
        await db.set_db_encryption(hashed)

        print("PASSWORD SET", hashed)

    def remove_encrypt_buttons(self) -> None:
        buttons: list[Button] = self.query("Button")
        label_container = self.query_one("#label_container", Center)
        label_container.remove()
        for button in buttons:
            if button.id in ["no_encrypt_db", "encrypt_db"]:
                button.remove()
                # button.disabled = True
            elif button.id in ["create_wallet", "from_mnemonic", "from_xpriv"]:
                button.disabled = False

    def on_button_pressed(self, event: Button.Pressed):
        event.stop()
        self.set_focus(None)

        # if event.button.id == "create_wallet":
        #     self.app.switch_screen("create_wallet")
        if event.button.id == "from_mnemonic":
            self.app.switch_screen("from_mnemonic")
        if event.button.id == "encrypt_db":
            self.app.push_screen(EncryptionPassword(), self.on_encryption_password_set)
            self.remove_encrypt_buttons()
        if event.button.id == "create_wallet":
            self.app.switch_screen("create_wallet")

        if event.button.id == "no_encrypt_db":
            self.remove_encrypt_buttons()

        # self.set_focus(None)



from textual.app import ComposeResult
from textual.containers import Grid
from textual.screen import ModalScreen
from textual.widgets import Button, Static


class SendTxOverlay(ModalScreen[bool]):
    DEFAULT_CSS = """
    SendTxOverlay {
        align: center middle;
    }

    SendTxOverlay > Grid {
        grid-size: 2;
        grid-rows: 1fr 1fr 1fr 3;
        padding: 0 1;
        grid-gutter: 1 2;
        width: 50%;
        height: 80%;
        border: thick $background 80%;
        background: $surface;
    }

    .info {
        border: tall $primary;
        column-span: 2;
        height: 1fr;
        width: 1fr;
        content-align: left middle;
        padding-left: 1;
    }
    #message {
        content-align: left top
    }

    SendTxOverlay > Grid > Button {
        width: 100%;
    }
    """

    def __init__(self, address: str, amount: float, message: str) -> None:
        self.address = address
        self.amount = amount
        self.message = message

        super().__init__()

    def compose(self) -> ComposeResult:
        yield Grid(
            Static(f"Send to: {self.address}", classes="info"),
            Static(f"Amount to send: {str(self.amount)} FLUX", classes="info"),
            Static(f"Message: {self.message}", classes="info", id="message"),
            Button("Confirm", variant="success", id="confirm"),
            Button("Discard", variant="error", id="discard"),
        )

    def on_mount(self) -> None:
        self.query_one("#discard", Button).focus()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        send = True if event.button.id == "confirm" else False
        self.dismiss(send)



import webbrowser

from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.screen import ModalScreen
from textual.widgets import Button, Label, Static


class ExplorerLabel(Label):
    def action_open_explorer_link(self, txid: str):
        base = "https://explorer.runonflux.io/tx"
        webbrowser.open(f"{base}/{txid}")


class TxSentOverlay(ModalScreen):
    def __init__(self, txid: str):
        self.txid = txid
        super().__init__()

    def on_mount(self) -> None:
        self.set_focus(None)

    def compose(self) -> ComposeResult:
        label = f"View on explorer: [@click=open_explorer_link('{self.txid}')]{self.txid}[/]"
        yield Container(
            Horizontal(
                ExplorerLabel(label, id="txid"),
                Container(
                    Button("X", variant="error", id="exit"), id="button_container"
                ),
            ),
            Static("Transaction Sent!", id="sent_message"),
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.dismiss()



from rich.text import Text
from textual.app import ComposeResult
from textual.reactive import var
from textual.widget import Widget
from textual.widgets import Button, Static

from textual import work
from textual.app import App
from rich.text import Text

from fluxwallet.db_new import Db, DbConfig

from decimal import Decimal

# to display

# Blockheight: 1446174 Hashrate: 1.8 MS/S Price 0.41 USD Marketcap: 130.1M USD 24Hr: ↑0.43%


class NetworkBar(Widget):
    DEFAULT_CSS = """
    # NetworkBar {
    #     margin-top: 1;
    #     layout: horizontal;
    #     background: $panel;
    #     height: 3;
    # }
    # NetworkBar>Static {
    #     content-align: center middle;
    #     border: tall $primary;
    #     width: 1fr;
    # }
    # NetworkBar>Button {
    #     dock: right;
    #     width: auto;
    #     min-width: 0;
    #     margin-right: 0;
    #     border: tall $primary;
    # }
    # NetworkBar>Button:focus {
    #     text-style: none;
    #     background: black 80%;
    # }
    """

    blockheight = var(0)
    hashrate = var(0.0)
    price = var(0.0)
    marketcap = var(0)
    delta_24hr = var(0.0)
    show_market_data = var(True)

    def __init__(
        self,
        blockheight: int = 0,
        hashrate: float = 0.0,
        price: float = 0.0,
        marketcap: float = 0.0,
        delta_24hr: float = 0.0,
        show_market_data: bool = True,
    ) -> None:
        self._blockheight = blockheight
        self._hashrate = hashrate
        self._price = price
        self._marketcap = marketcap
        self._delta_24hr = delta_24hr
        self._show_market_data = show_market_data

        self.db = Db()

        super().__init__()

    def on_mount(self) -> None:
        self.blockheight = self._blockheight
        self.hashrate = self._hashrate
        self.price = self._price
        self.marketcap = self._marketcap
        self.delta_24hr = self._delta_24hr
        self.show_market_data = self._show_market_data

    def compose(self) -> ComposeResult:
        yield Static(
            self.build_blockheight(), id="blockheight", classes="networkbar_item"
        )
        yield Static(self.build_hashrate(), id="hashrate", classes="networkbar_item")
        yield Button("\U0001F4B0", id="market_data_toggle")

        if self._show_market_data:
            yield Static(
                self.build_price(), id="price", classes="marketdata networkbar_item"
            )
            yield Static(
                self.build_marketcap(),
                id="marketcap",
                classes="marketdata networkbar_item",
            )
            yield Static(
                self.build_delta_24hr(),
                id="delta_24hr",
                classes="marketdata networkbar_item",
            )

    def build_blockheight(self) -> Text:
        return Text(f"Blockheight: {self._blockheight}")

    def build_hashrate(self) -> Text:
        return Text(f"Hashrate: {round(self._hashrate / 10**6, 2)} MS/s")

    def build_price(self) -> Text:
        # decimal will print the 3rd digit if its 0
        return Text(f"Price: {round(Decimal(self._price), 3)} USD")

    def build_marketcap(self) -> Text:
        return Text(f"Marketcap: {round(self._marketcap / 10**6, 2)}M USD")

    def build_delta_24hr(self) -> Text:
        UP_ARROW = "\U00002191"
        DOWN_ARROW = "\U00002193"
        positive_style = "bold green"
        negative_style = "bold red"

        if self._delta_24hr >= 0:
            symbol = UP_ARROW
            style = positive_style
        else:
            symbol = DOWN_ARROW
            style = negative_style

        delta = Text.assemble("24hr: ", (f"{symbol} {self._delta_24hr} %", style))

        return delta

    @work(group="set_db_show_market_data", exclusive=True)
    async def set_db_show_market_data(self, show_market_data: bool) -> None:
        async with self.db.get_session() as session:
            await session.merge(
                DbConfig(variable="show_market_data", value=str(int(show_market_data)))
            )
            await session.commit()

    def watch_blockheight(self, new_value: int) -> None:
        self._blockheight = new_value

        height = self.query_one("#blockheight", Static)
        height.update(self.build_blockheight())

    def watch_hashrate(self, new_value: float) -> None:
        self._hashrate = new_value

        hashrate = self.query_one("#hashrate", Static)
        hashrate.update(self.build_hashrate())

    def watch_price(self, new_value: float) -> None:
        self._price = new_value

        if self._show_market_data:
            price = self.query_one("#price", Static)
            price.update(self.build_price())

    def watch_marketcap(self, new_value: float) -> None:
        self._marketcap = new_value

        if self._show_market_data:
            marketcap = self.query_one("#marketcap", Static)
            marketcap.update(self.build_marketcap())

    def watch_delta_24hr(self, new_value: float) -> None:
        self._delta_24hr = new_value

        if self._show_market_data:
            delta_24hr = self.query_one("#delta_24hr", Static)
            delta_24hr.update(self.build_delta_24hr())

    def watch_show_market_data(self, new_value: bool) -> None:
        if self._show_market_data == new_value:
            return

        if not self._show_market_data and new_value:
            self.mount(
                Static(
                    self.build_price(), id="price", classes="marketdata networkbar_item"
                )
            )
            self.mount(
                Static(
                    self.build_marketcap(),
                    id="marketcap",
                    classes="marketdata networkbar_item",
                )
            )
            self.mount(
                Static(
                    self.build_delta_24hr(),
                    id="delta_24hr",
                    classes="marketdata networkbar_item",
                )
            )
        else:
            market_data = self.query(".marketdata")
            for widget in market_data:
                widget.remove()

        self.set_db_show_market_data(new_value)

        self._show_market_data = new_value

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.show_market_data = False if self.show_market_data == True else True


if __name__ == "__main__":

    class blahApp(App):
        def compose(self):
            self.bar = NetworkBar()
            yield self.bar

        def on_mount(self):
            self.bar.blockheight = 1446174
            self.bar.hashrate = 1801598
            self.bar.price = 0.41
            self.bar.marketcap = 129989111
            self.bar.delta_24hr = 0.47
            self.call_after_refresh(self.set_focus, None)

    app = blahApp()
    app.run()



from time import monotonic

from rich.text import Text
from textual.app import ComposeResult
from textual.events import MouseScrollDown, Focus, DescendantFocus
from textual.message import Message
from textual.reactive import var
from textual.widget import Widget
from textual.widgets import DataTable, Label, Static
from textual.containers import Center

from lucidwallet.widgets import TxLoading


class TransactionHistory(Widget):
    HEADER = ("Out/In", "Date", "Time", "Value", "Units")
    DEFAULT_CSS = """
    # TransactionHistory {
    #     width: 1fr;
    #     margin-top: 1;
    #     margin-bottom: 1;
    #     background: $panel;
    #     outline-top: tall $primary;
    #     outline-left: tall $primary;
    #     outline-right: tall $primary;
    #     layers: base notifications;
    # }
    # .top {
    #     outline-top: tall $primary;
    #     width: 100%;
    #     height: 3;
    #     margin: 0 1 0 1;
    #     content-align: center middle;
    #     background: $panel;
    # }
    # .bottom {
    #     outline-bottom: tall $primary;
    #     width: 100%;
    #     height: 2;
    #     margin: 0 1 0 1;
    #     padding-bottom: 1;
    #     background: $panel;
    # }
    """

    rows = var([])

    class ScrollDataTable(DataTable):
        def _on_focus(self, event: Focus) -> None:
            # hack to stop scrolling on DataTable click
            event.prevent_default()
            # self.has_focus = True
            self.refresh()
            self.post_message(DescendantFocus())

    class ScrollCenter(Center):
        class LazyLoadRequested(Message):
            ...

        def __init__(self, *args, **kwargs) -> None:
            self.last_lazyload = 0.0
            super().__init__(*args, **kwargs)

        def on_mouse_scroll_down(self, event: MouseScrollDown) -> None:
            """Sends a lazy load request if we're near the bottom of the screen.
            Uses a 300ms debouce on lazy load requests

            Args:
                event (MouseScrollDown): unused scroll event
            """
            now = monotonic()
            if self.last_lazyload + 0.3 < now:
                self.last_lazyload = now

                if self.max_scroll_y == 0:
                    return

                match self.max_scroll_y:
                    case x if x < 50:
                        ratio = 0.4
                    case x if x >= 50 and x < 100:
                        ratio = 0.6
                    case x if x >= 100 < 200:
                        ratio = 0.75
                    case _:
                        ratio = 0.80

                if self.scroll_y / self.max_scroll_y > ratio:
                    self.post_message(self.LazyLoadRequested())

    class TxOverlayRequested(Message):
        def __init__(self, txid: str):
            self.txid = txid
            super().__init__()

    def __init__(self, rows: list[tuple[str | Text]]) -> None:
        self._rows = rows
        self.selected_row: str | None = None
        self.loading = False
        self.loading_widget = TxLoading(id="tx_loading")
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Label("Previous Transactions", classes="top")

        with self.ScrollCenter():
            yield self.ScrollDataTable()

        yield Label("", classes="bottom")

    def on_mount(self) -> None:
        table = self.query_one("ScrollDataTable", self.ScrollDataTable)

        for col in self.HEADER:
            table.add_column(col, key=col)

        self.rows = self._rows
        table.cursor_type = "row"

    def on_data_table_row_selected(self, event: DataTable.RowSelected):
        event.stop()

        if not event.row_key.value:
            return

        if self.selected_row != event.row_key.value:
            self.selected_row = event.row_key.value
            return

        self.post_message(self.TxOverlayRequested(event.row_key.value))

        # could add a timing element here? Like 100ms?
        # send message with txdata to parent

    def clear_table(self) -> None:
        self.rows = []

    async def set_loading(self) -> None:
        if not self.loading:
            self.loading = True
            self.disabled = True
            self.call_after_refresh(self.mount, self.loading_widget)

    def unset_loading(self) -> None:
        if self.loading:
            self.loading = False
            self.disabled = False
            self.call_after_refresh(self.loading_widget.remove)

    def scroll(self, y: float) -> None:
        scroll_center = self.query_one("ScrollCenter", self.ScrollCenter)
        self.call_after_refresh(scroll_center.scroll_to, y=y, animate=False)

    def update_date_time(self, txid: str, date: str, time: str) -> None:
        table = self.query_one("ScrollDataTable", self.ScrollDataTable)
        print("UPDATING TIME", time)
        table.update_cell(txid, "Date", date)
        table.update_cell(txid, "Time", time)

    def watch_rows(self, new_value: list[tuple]):
        table = self.query_one("ScrollDataTable", self.ScrollDataTable)
        self._rows = new_value

        if not self._rows:
            table.clear()

        for row in self._rows:
            table.add_row(*row[1:], key=row[0])

        table.sort("Date", "Time", reverse=True)



import asyncio
import socketio

from functools import partial

from fluxwallet.db_new import Db, DbAddressBook, DbConfig
from fluxwallet.keys import HDKey
from fluxwallet.values import Value
from fluxwallet.wallet import Wallet, WalletKey, WalletTransaction, WalletError
from importlib_metadata import version
from rich.console import RenderableType
from sqlalchemy import select
from textual import on, work
from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.message import Message
from textual.reactive import reactive
from textual.screen import Screen
from textual.widgets import Footer, Header, Input, Static, Switch

from lucidwallet.datastore import FluxWalletDataStore, ScanType, Timer
from lucidwallet.events import Event
from lucidwallet.screens import (
    AddressBook,
    ImportKeyToWallet,
    SendTxOverlay,
    SignMessageOverlay,
    TxOverlay,
    TxSentOverlay,
)
from lucidwallet.widgets import Send, TopBar, NetworkBar, TransactionHistory

# from textual.css.query import NoMatches


NAV_LINKS = """
[@click="follow_nav('address_book')"]Address Book[/]
[@click="follow_nav('create_wallet')"]Create Wallet[/]
[@click="follow_nav('import_wallet')"]Import Wallet[/]
[@click="follow_nav('import_key')"]Import Key[/]
[@click="follow_nav('encrypt_database')"]Encrypt Database Keys[/]

"""


class Navigation(Static):
    class NavOpened(Message):
        def __init__(self, nav_target: str | None = None) -> None:
            self.nav_target = nav_target
            super().__init__()

    class AddressBookUpdated(Message):
        ...

    def action_follow_nav(self, nav_item: str) -> None:
        self.post_message(self.NavOpened())

        match nav_item:
            case "address_book":
                # do this with post message as currently will reload txs
                self.app.push_screen(AddressBook(), self.on_address_book_update)
            case "import_wallet":
                self.app.push_screen("from_mnemonic")
            case "import_key":
                self.post_message(self.NavOpened(nav_item))
            case "encrypt_database":
                print("WOULD ENCRYPT HERE, NOT BUILT YET")
            case "create_wallet":
                self.app.push_screen("create_wallet")

    def on_address_book_update(self, selected: str | None):
        # shouldn't pass thought the param here from AB.
        self.post_message(self.AddressBookUpdated())


class Version(Static):
    def render(self) -> RenderableType:
        return f"[b]v{version('fluxwallet')}"


class Title(Static):
    pass


class OptionGroup(Container):
    pass


class DarkSwitch(Horizontal):
    def compose(self) -> ComposeResult:
        yield Switch(value=self.app.dark)
        yield Static("Dark mode toggle", classes="label")

    def on_mount(self) -> None:
        self.watch(self.app, "dark", self.on_dark_change, init=False)

    def on_dark_change(self) -> None:
        self.query_one(Switch).value = self.app.dark

    def on_switch_changed(self, event: Switch.Changed) -> None:
        self.app.dark = event.value


class Sidebar(Container):
    def compose(self) -> ComposeResult:
        yield Title("LucidWallet Menu")
        yield OptionGroup(Navigation(NAV_LINKS), Version())
        yield DarkSwitch()

    def hide(self) -> None:
        self.add_class("-hidden")

    @on(Navigation.NavOpened)
    def on_nav(self):
        self.hide()


class WalletLanding(Screen):
    # CSS_PATH = "wallet.css"
    TITLE = "LucidWallet"
    BINDINGS = [("ctrl+z", "toggle_sidebar", "Sidebar")]

    show_sidebar = reactive(False)

    class TxSentMessage(Message):
        def __init__(self, txid: str = "", error: bool = None, error_msg: str = ""):
            self.txid = txid
            self.error = error
            self.error_msg = error_msg
            super().__init__()

    class NetworkScanned(Message):
        def __init__(
            self,
            wallet_name: str,
            network_name: str,
            new_transactions: int,
            scan_type: ScanType,
        ) -> None:
            self.wallet_name = wallet_name
            self.network_name = network_name
            self.new_transactions = new_transactions
            self.scan_type = scan_type
            super().__init__()

    def __init__(
        self,
        initial_wallet: Wallet,
        initial_wallet_networks: list[str],
        wallets: list[str],
    ):
        self.initial_wallet = initial_wallet
        self.initial_wallet_networks = initial_wallet_networks

        self.datastore = FluxWalletDataStore(
            wallet_names=set(wallets),
            # scan_timer=Timer(self.periodic_scan, 30, run_on_start=True),
        )
        self.tx_events: asyncio.Queue = asyncio.Queue()
        self.update_dom_on_resume = True
        self.sio = socketio.AsyncClient()

        super().__init__()

    def compose(self) -> ComposeResult:
        yield Header(show_clock=True)
        yield NetworkBar(
            blockheight=self.app.config.network_data.blockheight,
            hashrate=self.app.config.network_data.hashrate,
            price=self.app.config.network_data.price,
            marketcap=self.app.config.network_data.marketcap,
            delta_24hr=self.app.config.network_data.delta_24hr,
            show_market_data=self.app.config.show_market_data,
        )
        # fix this, just pass in defaults and tidy it up after.
        yield TopBar(
            selected_wallet=self.initial_wallet.name,
            selected_network="flux",
            networks=self.initial_wallet_networks,
            wallets=list(self.datastore.wallet_names),
        )
        with Horizontal():
            yield Send()
            yield TransactionHistory([])

        yield Footer()

    @work(group="get_tx_from_db_worker")
    async def get_tx_from_db_worker(
        self, force: bool = False, latest_only: bool = False, limit: int = 60
    ) -> None:
        await self.datastore.get_transactions_from_db(
            self.tx_events, force=force, latest_only=latest_only, limit=limit
        )

    async def on_explorer_info(self, data: dict) -> None:
        print(data)
        # example = {
        #     "info": {
        #         "version": 6020050,
        #         "protocolversion": 170018,
        #         "walletversion": 60000,
        #         "blocks": 1446160,
        #         "timeoffset": 0,
        #         "connections": 16,
        #         "proxy": "",
        #         "difficulty": 27317.79969724339,
        #         "testnet": False,
        #         "relayfee": 1e-06,
        #         "errors": "",
        #         "network": "livenet",
        #         "reward": 3750000000,
        #     },
        #     "miningInfo": {"difficulty": 26425.96416617444, "networkhashps": 1817155},
        #     "supply": "395459750",
        # }
        blockheight = data["info"]["blocks"]
        network_hash = data["miningInfo"]["networkhashps"]

        network_bar = self.query_one("NetworkBar", NetworkBar)
        network_bar.hashrate = network_hash
        network_bar.blockheight = blockheight

        self.periodic_scan_worker(blockheight=blockheight)

    async def on_explorer_market_info(self, data: dict) -> None:
        # example = {
        #     "price": 0.4115384952910444,
        #     "price_btc": 1.398345629016484e-05,
        #     "market_cap_usd": 130147424,
        #     "total_volume_24h": 1317329.0410391006,
        #     "delta_24h": 0.8,
        # }
        print(data)
        network_bar = self.query_one("NetworkBar", NetworkBar)
        network_bar.price = data["price"]
        network_bar.marketcap = data["market_cap_usd"]
        network_bar.delta_24hr = data["delta_24h"]

    async def on_unmount(self) -> None:
        await self.sio.disconnect()

    async def on_mount(self) -> None:
        print("WALLET LANDING ON MOUNT")
        self.monitor_tx_history()

        # just pass in wallet name to WalletLanding, don't open wallet
        await self.datastore.set_current_wallet(self.initial_wallet.name)
        # self.datastore.start_scan_timer()
        await self.set_dom_spend_details()
        # self.get_tx_from_db_worker()

        # try/except. If this fails... fallback to polling scantimer
        # websocket stuff needs to be on wallet, not here
        self.sio.on("info", self.on_explorer_info)
        self.sio.on("markets_info", self.on_explorer_market_info)
        await self.sio.connect(
            "https://explorer.runonflux.io", transports=["websocket"]
        )
        await self.sio.emit("subscribe", "inv")

        self.periodic_scan_worker(self.app.config.network_data.blockheight)

        self.initial_wallet = None
        self.initial_wallet_networks = None

        self.db = await Db.start()
        current_address_book = await self.fetch_adress_book()

        send = self.query_one("Send", Send)
        send.update_address_book("", current_address_book)

        self.mount(Sidebar(classes="-hidden"))

    def worker_running(self, worker_name: str) -> bool:
        return bool(next(filter(lambda x: x.group == worker_name, self.workers), None))

    # should this be on the datastore
    @work(group="set_db_last_used_wallet", exclusive=True)
    async def set_db_last_used_wallet(self, wallet_name: str) -> None:
        async with self.db.get_session() as session:
            await session.merge(
                DbConfig(variable="last_used_wallet", value=wallet_name)
            )
            await session.commit()

    async def fetch_adress_book(self) -> dict[str, str]:
        async with self.db.get_session() as session:
            res = await session.scalars(select(DbAddressBook))
            address_book = res.all()

        return {x.name: x.address for x in address_book}

    def set_dom_wallet_details(self) -> None:
        topbar = self.query_one("TopBar", TopBar)

        current_wallet = self.datastore.current_wallet
        current_network = self.datastore.current_network
        wallets = self.datastore.get_known_wallets()
        networks = self.datastore.get_current_wallet_networks()

        topbar.set_wallet_options(current_wallet, current_network, wallets, networks)

    async def set_dom_spend_details(self) -> None:
        receive_address, balance = await self.datastore.get_current_wallet_spend_info()
        topbar = self.query_one("TopBar", TopBar)
        # temp until I fix api and remove conditional in topbar
        topbar.selected_network = self.datastore.current_network
        topbar.receive_address = receive_address
        topbar.balance = balance

    @work(group="tx_history_updater")
    async def monitor_tx_history(self) -> None:
        tx_history = self.query_one("TransactionHistory", TransactionHistory)

        while True:
            event: Event = await self.tx_events.get()

            match event.type:
                case Event.EventType.NewRows:
                    tx_history.rows = event.rows
                case Event.EventType.ClearTable:
                    tx_history.clear_table()
                case Event.EventType.ScanningStart:
                    await tx_history.set_loading()
                case Event.EventType.ScanningEnd:
                    tx_history.unset_loading()
                case Event.EventType.Scroll:
                    if event.scroll_height:
                        print("scrolling to", event.scroll_height)
                        tx_history.scroll(y=event.scroll_height)

    def tx_history_dom_reload(self) -> None:
        print("TXHISTORY DOM RELOAD")
        self.tx_events.put_nowait(Event(type=Event.EventType.ClearTable))

        self.workers.cancel_group(self, "get_tx_from_db_worker")

        # bit of a hack, keep getting duplicate errors, so only update once finished
        # scanning. Should fix
        if not self.datastore.is_current_network_scanning():
            self.get_tx_from_db_worker()

    async def update_send_address_book(self, selected: str | None = None):
        current_address_book = await self.fetch_adress_book()
        send = self.query_one("Send", Send)
        send.update_address_book(selected, current_address_book)

    def action_toggle_sidebar(self) -> None:
        sidebar = self.query_one("Sidebar", Sidebar)
        # self.set_focus(None)
        if sidebar.has_class("-hidden"):
            sidebar.remove_class("-hidden")
        else:
            if sidebar.query("*:focus"):
                self.screen.set_focus(None)
            sidebar.add_class("-hidden")

    @work(group="set_current", exclusive=True)
    async def set_current_and_update_dom(
        self, *, wallet: str | None = None, network: str | None = None
    ) -> None:
        dt = self.query_one("ScrollCenter", TransactionHistory.ScrollCenter)
        self.datastore.store_current_network_scroll_height(dt.scroll_y)
        print("SET CURRENT")
        send = self.query_one("Send", Send)

        if wallet:
            if self.datastore.current_network != "flux":
                send.unmount_disabled()
            await self.datastore.set_current_wallet(wallet)

        if network:
            if network != "flux":
                send.mount_disabled()
            else:
                send.unmount_disabled()
            self.datastore.set_current_network(network)

        await self.update_dom()
        # await self.datastore.reset_timer()

    @on(TopBar.WalletSelected)
    async def on_topbar_wallet_selected(self, event: TopBar.WalletSelected) -> None:
        if self.datastore.current_wallet == event.wallet:
            return

        # this is a worker now - db call.
        self.set_db_last_used_wallet(event.wallet)

        self.set_current_and_update_dom(wallet=event.wallet)

    @on(TopBar.NetworkSelected)
    async def on_topbar_network_selected(self, event: TopBar.NetworkSelected) -> None:
        if self.datastore.current_network == event.network:
            return

        self.set_current_and_update_dom(network=event.network)

    @on(Send.AddressBookUpdateRequested)
    def on_address_book_update(self, event: Send.AddressBookUpdateRequested) -> None:
        self.app.push_screen(
            AddressBook(event.address, return_nickname=True),
            self.update_send_address_book,
        )
        self.update_dom_on_resume = False

    @on(Navigation.AddressBookUpdated)
    async def on_nav_address_book_updated(self, event: Navigation.AddressBookUpdated):
        await self.update_send_address_book()

    @on(TransactionHistory.TxOverlayRequested)
    async def on_tx_overlay_requested(
        self, event: TransactionHistory.TxOverlayRequested
    ):
        tx = self.datastore.get_transaction(event.txid)
        await tx.sync()
        self.app.push_screen(TxOverlay(tx))
        self.update_dom_on_resume = False

    @on(Send.MaxAmountRequested)
    def set_max_amount(self, event: Send.MaxAmountRequested):
        event.stop()

        amount = self.query_one("#amount", Input)
        balance = self.query_one("TopBar", TopBar).balance
        # fix this - get the fee
        amount.value = str(max(0, float(Value(balance) - Value("225 sat"))))
        amount.focus()

    def tx_overlay_callback(
        self, address: str, amount: float, message: str, send: bool
    ):
        if not send:
            return

        self.send_transaction(address, amount, message)

        self.query_one("Send", Send).clear_fields()

    @on(Send.SendTxRequested)
    async def on_send_tx_requested(self, event: Send.SendTxRequested):
        balance = await self.datastore.current_network_balance()
        # Do this better, in fact, sort out the fee in FluxWallet too
        if float(event.amount) > float(Value(balance) - Value("225 sat")):
            self.notify(
                f"Not enough funds, {event.amount} + 225 Dibis > {balance}", timeout=5
            )
            return

        tx_confirm_callback = partial(
            self.tx_overlay_callback, event.address, event.amount, event.message
        )
        self.app.push_screen(
            SendTxOverlay(event.address, event.amount, event.message),
            tx_confirm_callback,
        )
        self.update_dom_on_resume = False

    @on(TxSentMessage)
    async def on_tx_sent(self, event: TxSentMessage):
        self.app.push_screen(TxSentOverlay(event.txid))
        self.update_dom_on_resume = False
        # self.run_worker(self.sync_wallet())
        await self.set_dom_spend_details()

    @on(TopBar.SignMessageRequested)
    async def on_sign_message_requested(self, event: TopBar.SignMessageRequested):
        wallet = self.datastore.get_current_wallet()
        wallet_key = await wallet.get_key(account_id=999, network="bitcoin")
        key = wallet_key.key()
        self.app.push_screen(SignMessageOverlay(key.private_byte))
        self.update_dom_on_resume = False

    @on(ImportKeyToWallet.KeyImportRequested)
    def on_key_import_requested(self, event: ImportKeyToWallet.KeyImportRequested):
        ...

    @on(Navigation.NavOpened)
    def on_nav_opened(self, event: Navigation.NavOpened) -> None:
        match event.nav_target:
            case "import_key":
                self.app.push_screen(
                    ImportKeyToWallet(
                        self.datastore.current_wallet,
                        self.datastore.wallet_names,
                        self.datastore.current_network,
                        self.datastore.get_current_wallet_networks(),
                    ),
                    self.import_key_callback,
                )
                self.update_dom_on_resume = False

    @on(TransactionHistory.ScrollCenter.LazyLoadRequested)
    def on_lazyload_requested(self) -> None:
        if not self.worker_running("get_tx_from_db_worker"):
            self.get_tx_from_db_worker(force=True)

    @on(NetworkScanned)
    async def on_wallet_scanned(self, event: NetworkScanned):
        if (
            event.wallet_name != self.datastore.current_wallet
            and event.network_name != self.datastore.current_network
        ):
            print("current wallet / network not same as scanned network")
            return

        print("NEW TXS", event.new_transactions)

        if event.scan_type != ScanType.PERIODIC:
            # maybe this needs a worker... (db call for balance)
            await self.update_dom()
            return

        if event.new_transactions:
            await self.set_dom_spend_details()
            limit = min(60, event.new_transactions)
            self.get_tx_from_db_worker(force=True, latest_only=True, limit=limit)
            self.notify("New Tx received", timeout=10)

    @on(TopBar.RescanAllRequested)
    async def on_rescan_all_requested(self) -> None:
        # self.full_scan_required = True
        self.rescan_wallet()

    async def on_screen_resume(self) -> None:
        # hack until I can think of how to do it better
        if self.update_dom_on_resume:
            print("UPDATING DOM ON RESUME")
            await self.update_dom()
        self.update_dom_on_resume = True

    @work(group="import_key_and_scan")
    async def import_key_and_scan(
        self, wallet_name: str, network_name: str, private_key: str
    ) -> None:
        key = HDKey(private_key, network=network_name)

        if (
            wallet_name != self.datastore.current_wallet
            or network_name != self.datastore.current_network
        ):
            await self.datastore.set_current_wallet(wallet_name)
            self.datastore.set_current_network(network_name)
            await self.update_dom()
            # await self.datastore.reset_timer()

        wallet = self.datastore.get_current_wallet()

        wallet_key = await wallet.import_key(key, network=network_name)

        if wallet_key:
            # self.tx_events.put_nowait(Event(type=Event.EventType.ClearTable))
            self.key_scan(wallet_key)
        else:
            self.notify("Key already imported")

    async def import_key_callback(self, result: tuple | None) -> None:
        if not result:
            return

        wallet_name, network_name, private_key = result
        self.import_key_and_scan(wallet_name, network_name, private_key)

    @work()
    async def send_transaction(self, address: str, amount: float, message: str):
        value = Value(amount, network="flux")

        wallet = self.datastore.get_current_wallet()

        try:
            wt: WalletTransaction = await wallet.transaction_create(
                [(address, value)], message=message
            )
        except WalletError as e:
            self.notify(e)
            return

        # maybe to_thread this. Benchmarked at 5ms. Fine.
        wt.sign()
        await wt.send()

        if wt.pushed:
            message = self.TxSentMessage(wt.tx.txid)
            self.datastore.add_unconfirmed_tx(wt)
            self.get_tx_from_db_worker(force=True, latest_only=True, limit=1)
        else:
            message = self.TxSentMessage(error=True, error_msg=wt.error)

        self.post_message(message)

    async def update_dom(self) -> None:
        current_network_scanning = self.datastore.is_current_network_scanning()

        if self.datastore.tx_history_scanning and not current_network_scanning:
            self.datastore.tx_history_scanning = False
            self.tx_events.put_nowait(Event(type=Event.EventType.ScanningEnd))

        if not self.datastore.tx_history_scanning and current_network_scanning:
            self.datastore.tx_history_scanning = True
            self.tx_events.put_nowait(Event(type=Event.EventType.ScanningStart))

        await self.set_dom_spend_details()
        self.set_dom_wallet_details()

        # if self.datastore.is_current_network_scanning():
        #     self.tx_events.put_nowait(Event(type=Event.EventType.ScanningStart))

        self.tx_history_dom_reload()

    async def new_wallet_created(self, wallet: Wallet) -> None:
        self.datastore.add_known_wallet(wallet.name)

        await self.datastore.set_current_wallet(wallet)

        self.full_wallet_scan()

    async def update_unconfirmed_txs(self) -> None:
        # fix this whole thing
        for wt in self.datastore.unconfirmed_txs:
            await wt.sync()

            if wt.status == "confirmed":
                date = wt.tx.date.date()
                time = wt.tx.date.strftime("%H:%M:%S")

                tx_history = self.query_one("TransactionHistory", TransactionHistory)
                tx_history.update_date_time(wt.tx.txid, date, time)
                self.notify(f"Tx @ {time} confirmed", timeout=10)

        self.datastore.unconfirmed_txs = [
            x for x in self.datastore.unconfirmed_txs if not x.status == "confirmed"
        ]

    def rescan_wallet(self):
        print("SCAN WALLET WORKER")
        # debounce
        if not self.datastore.scan_for_current_network_required(ScanType.FULL_WALLET):
            self.notify("Scanned already")
            return

        self.full_wallet_scan()

    # types of scans
    #
    # periodic, global - shared between all wallets. rerun (reset) on wallet change. Skip if wallet
    # wallet or key scan running for this wallet.
    #
    # new wallet - initial wallet scan
    #
    # new key - initial key scan
    #
    # rescan wallet (same as new wallet scan) if new wallet scan running, skip

    @work(group="periodic_scan_worker", exclusive=True)
    async def periodic_scan_worker(self, blockheight: int | None = None) -> None:
        # allow a few seconds for the api to update the new block
        await asyncio.sleep(5)
        await self.periodic_scan(blockheight=blockheight)

    async def periodic_scan(self, blockheight: int | None = None) -> None:
        print("PERIODIC SCAN")

        # if not self.datastore.scan_for_current_network_required(ScanType.PERIODIC):
        #     print("Scanned within the last 15 sec... returning")
        #     return

        wallet = self.datastore.get_current_wallet()
        network = self.datastore.get_current_network()

        if self.worker_running("key_scan") or self.worker_running("full_wallet_scan"):
            print("IN PERIODIC SCAN, OTHER SCANS RUNNING.... RETURNING")
            return

        await self.scan_network(
            wallet, network, scan_type=ScanType.PERIODIC, blockheight=blockheight
        )

    @work(group="key_scan")
    async def key_scan(self, key: WalletKey) -> None:
        self.datastore.reset_current_wallet_datastore()
        wallet = self.datastore.get_current_wallet()
        network = self.datastore.get_current_network()

        await self.scan_network(
            wallet, network=network, key=key, scan_type=ScanType.NEW_KEY
        )

    @work(group="full_wallet_scan")
    async def full_wallet_scan(self) -> None:
        # await self.datastore.reset_timer(run_on_start=False)
        self.datastore.reset_current_wallet_datastore()
        # self.tx_events.put_nowait(Event(type=Event.EventType.ClearTable))

        wallet = self.datastore.get_current_wallet()
        network = self.datastore.get_current_network()

        await self.scan_network(wallet, network, scan_type=ScanType.FULL_WALLET)

    def set_scanning_for_network(self, wallet_name: str, network_name: str) -> None:
        self.datastore.set_scanning_for_network(wallet_name, network_name)

        if (
            wallet_name == self.datastore.current_wallet
            and network_name == self.datastore.current_network
            and not self.datastore.tx_history_scanning
        ):
            print("SETTING TX DOM SCANNING")
            self.datastore.tx_history_scanning = True
            self.tx_events.put_nowait(Event(type=Event.EventType.ScanningStart))

    def unset_scanning_for_network(self, wallet_name: str, network_name: str) -> None:
        self.datastore.unset_scanning_for_network(wallet_name, network_name)

        if (
            wallet_name == self.datastore.current_wallet
            and network_name == self.datastore.current_network
            and self.datastore.tx_history_scanning
        ):
            self.datastore.tx_history_scanning = False
            self.tx_events.put_nowait(Event(type=Event.EventType.ScanningEnd))

    async def scan_network(
        self,
        wallet: Wallet,
        network: str,
        scan_type: ScanType,
        *,
        key: WalletKey | None = None,
        blockheight: int | None = None
        # rescan_used: bool = False,
    ) -> None:
        change = 0
        print("SCAN Network", network)

        rescan_used = False

        if not scan_type == ScanType.PERIODIC:
            self.set_scanning_for_network(wallet.name, network)

        if scan_type == ScanType.FULL_WALLET:
            rescan_used = True

        if self.datastore.is_network_first_scan(wallet.name, network) or rescan_used:
            self.datastore.set_network_scanned(wallet.name, network)
            change = None

        if key:
            new_txids = await wallet.scan_key(key)
        else:
            new_txids = await wallet.scan(
                change=change,
                rescan_used=rescan_used,
                network=network,
                blockcount=blockheight,
            )

        await self.update_unconfirmed_txs()

        if not scan_type == ScanType.PERIODIC:
            self.unset_scanning_for_network(wallet.name, network)

        if not scan_type == ScanType.NEW_KEY:
            self.datastore.set_last_scanned_for_network(wallet.name, network, scan_type)

        self.post_message(
            self.NetworkScanned(
                wallet.name,
                network,
                scan_type=scan_type,
                new_transactions=len(new_txids),
            )
        )



from typing import Iterable

from textual import on
from textual.app import App, ComposeResult
from textual.containers import Center, Horizontal
from textual.messages import Message
from textual.reactive import reactive
from textual.validation import Function, Length, Number
from textual.widget import Widget
from textual.widgets import Button, Input, Label, Select

# from .top_bar import TopBar

# address_book = {
#     "gravywallet": "t1XvGeQCfYMhfagYb3GmbKRajNEjpPRZHkB",
#     "Anwen": "t1CvGeCCfYMhfagYAba4Re3ajNEjpPRZJ74",
# }


class Send(Widget):
    DEFAULT_CSS = """
    Send {
        height: 100%;
        width: 1fr;
        min-width: 60;
        # align: center middle;
        outline: tall $primary;
        margin-top: 1;
        margin-bottom: 1;
        background: $panel;
    }
    Send>Horizontal>Input {
        width: 2fr;
        border: tall $primary;
        # margin: 1;
    }
    Send>Center {
        margin-left: 1;
        margin-top: 1;
    }
    #save {
        margin: 0;
        padding: 0;
        height: auto;
        min-width: 0;
        width: auto;
    }
    #save:focus {
        text-style: none;
        outline-top: tall black;
        background: black 80%;
    }
    #buttons {
        margin-left: 1;
        margin-top: 1;
        align: center middle;
    }
    #all_funds {
        margin: 0;
        padding-right: 1;
        margin-left: 1;
        outline-right: tall black;
        min-width: 0;
    }
    #all_funds:focus {
        text-style: none;
        outline-top: tall black;
        background: black 80%;
    }
    Send>Horizontal>Button {
        margin: 1;

    }
    Send>Horizontal>Select>SelectCurrent {
        border: tall $primary;
        max-height: 3;
    }
    Send>Horizontal>Select>SelectCurrent Static#label {
       overflow: hidden;
    }
    """

    sendto_address = reactive("")
    sendto_amount = reactive(0.0)
    send_enabled = reactive(False)
    save_enabled = reactive(False)

    class SendTxRequested(Message):
        def __init__(self, address: str, amount: float, message: str):
            self.address = address
            self.amount = amount
            self.message = message
            super().__init__()

    class MaxAmountRequested(Message):
        ...

    class AddressBookUpdateRequested(Message):
        def __init__(self, address: str):
            self.address = address
            super().__init__()

    def __init__(self, address_book: dict[str, str] = {}):
        self.address_book = address_book
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Input(
            placeholder="Address",
            id="address",
            # validators=[
            #     Function(function=self.validate_address_input),
            # ],
        )
